# -*- coding: utf-8 -*-

__version__ = '1.0.0'
program_name = 'bio_hansel'
program_summary = 'Subtype Salmonella enterica genomes using 33bp k-mer typing schemes.'
program_desc = program_summary + '''
Includes schemes for Heidelberg and Enteritidis subtyping.
Developed by Geneviève Labbé, James Robertson, Peter Kruczkiewicz, Chad R. Laing, Kim Ziebell, Marisa Rankin, Aleisha R. Reimer, Lorelee Tschetter, Gary Van Domselaar, Eduardo N. Taboada, Sadjia Bekal, Kimberley A. MacDonald, Linda Hoang, Linda Chui, Danielle Daignault, Durda Slavic, Frank Pollari, E. Jane Parmley, Philip Mabon, Elissa Giang, Lok Kan Lee, Jonathan Moffat, Joanne MacKinnon, Benjamin M. Hetman, Roger Johnson, John H.E. Nash.
'''



from typing import List, Callable, Tuple

from pandas import DataFrame

from ..subtyping_params import SubtypingParams
from ..quality_check.quality_check_functions import check_missing_tiles, does_subtype_result_exist, \
    check_mixed_subtype, check_intermediate_subtype, is_missing_target_sites, is_missing_downstream_targets
from ..quality_check.const import FAIL_MESSAGE, WARNING_MESSAGE
from ..subtype import Subtype
import logging


QC_FUNCS = \
[
    check_missing_tiles,
    check_mixed_subtype,
    is_missing_target_sites,
    is_missing_downstream_targets,
    check_intermediate_subtype,
] # type: List[Callable[[Subtype, DataFrame, SubtypingParams], Tuple[str, str]]]


def perform_quality_check(st: Subtype, df: DataFrame, subtyping_params: SubtypingParams):
    """ Driver method to call all quality checking functions and handle their responses.
    Note:
            This is the driver method for the quality check module. Every method within the QC_FUNCS list will be run
            with parameters ( SUBTYPE, DATAFRAME ). If a quality check module returns something other than None, then
            an Error, or Warning has occured.

    Args:
            :param st: Subtyping results.
            :param df: DataFrame containing subtyping results.

    Returns:
            None, modifies the subtype with the result.
    """
    logging.debug("Performing Quality Checking")
    overall_qc_status = 'PASS'
    messages = []

    if does_subtype_result_exist(st) is False:
        logging.warning("QC: Quality checking not run, subtype result did not exist.")
        st.qc_status = 'FAIL'
        st.qc_message = 'FAIL: Subtype does not exist, quality checking was not run.'
        return None

    for func in QC_FUNCS:
        # Calls run_method to check that the qc function takes a Subtype, returns Tuple[Optional[str], Optional[str]]
        status, message = func(st, df, subtyping_params)
        if status is None:
            # If quality check function passes, move on to the next.
            continue
        messages.append('{}: {}'.format(status, message))
        if status is FAIL_MESSAGE:
            overall_qc_status = FAIL_MESSAGE
        elif overall_qc_status != FAIL_MESSAGE and status == WARNING_MESSAGE:
            overall_qc_status = WARNING_MESSAGE

    st.qc_status = overall_qc_status
    st.qc_message = ' | '.join(messages)
    logging.debug("QC: Finished!")



# -*- coding: utf-8 -*-
from typing import List, Optional

import attr

@attr.s
class Subtype(object):
    sample = attr.ib(validator=attr.validators.instance_of(str))
    file_path = attr.ib(validator=attr.validators.instance_of(str))
    scheme = attr.ib(validator=attr.validators.instance_of(str))
    scheme_version = attr.ib(default=None, validator=attr.validators.optional(attr.validators.instance_of(str)))
    subtype = attr.ib(default=None, validator=attr.validators.optional(attr.validators.instance_of(str)))
    non_present_subtypes = attr.ib(default=None) # type: Optional[List[str]]
    all_subtypes = attr.ib(default=None, validator=attr.validators.optional(attr.validators.instance_of(str)))
    inconsistent_subtypes = attr.ib(default=None, validator=attr.validators.optional(attr.validators.instance_of(str)))
    tiles_matching_subtype = attr.ib(default=None, validator=attr.validators.optional(attr.validators.instance_of(str)))
    negative_tiles_matching_subtype = attr.ib(default=None, validator=attr.validators.optional(attr.validators.instance_of(str)))
    are_subtypes_consistent = attr.ib(default=True, validator=attr.validators.instance_of(bool))
    n_tiles_matching_all = attr.ib(default=0, validator=attr.validators.instance_of(int))
    n_tiles_matching_positive = attr.ib(default=0, validator=attr.validators.instance_of(int))
    n_tiles_matching_negative = attr.ib(default=0, validator=attr.validators.instance_of(int))
    n_tiles_matching_subtype = attr.ib(default=0, validator=attr.validators.instance_of(int))
    n_tiles_matching_all_expected = attr.ib(default=0, validator=attr.validators.instance_of(int))
    n_tiles_matching_positive_expected = attr.ib(default=0, validator=attr.validators.instance_of(int))
    n_tiles_matching_negative_expected = attr.ib(default=0, validator=attr.validators.instance_of(int))
    n_tiles_matching_subtype_expected = attr.ib(default=0, validator=attr.validators.instance_of(int))
    n_negative_tiles_matching_subtype_expected = attr.ib(default=0, validator=attr.validators.instance_of(int))
    qc_status = attr.ib(default=None, validator=attr.validators.optional(attr.validators.instance_of(str)))
    qc_message = attr.ib(default=None, validator=attr.validators.optional(attr.validators.instance_of(str)))
    scheme_subtype_counts = attr.ib(default=None)



# -*- coding: utf-8 -*-
import logging
import numpy as np

#: set: valid IUPAC nucleotide characters for checking FASTA format
VALID_NUCLEOTIDES = {'A', 'a',
                     'C', 'c',
                     'G', 'g',
                     'T', 't',
                     'R', 'r',
                     'Y', 'y',
                     'S', 's',
                     'W', 'w',
                     'K', 'k',
                     'M', 'm',
                     'B', 'b',
                     'D', 'd',
                     'H', 'h',
                     'V', 'v',
                     'N', 'n',
                     'X', 'x', }  # X for masked nucleotides

NT_SUB = {x: y for x, y in zip('acgtrymkswhbvdnxACGTRYMKSWHBVDNX', 'tgcayrkmswdvbhnxTGCAYRKMSWDVBHNX')}


def revcomp(s):
    """Reverse complement nucleotide sequence

    Args:
        s (str): nucleotide sequence

    Returns:
        str: reverse complement of `s` nucleotide sequence
    """
    return ''.join([NT_SUB[c] for c in s[::-1]])


def extend_subj_match_vec(df):
    """
    Get the extended clipped (clamped) start and end subject sequence indices

    Also get whether each match needs to be reverse complemented and whether each extended match would be truncated by
    the end of the subject sequence.

    Args:
        df (pandas.DataFrame): blastn results dataframe

    Returns:
        int pandas.Series: extended and clipped start indices
        int pandas.Series: extended and clipped end indices
        bool pandas.Series: does extracted seq need to be reverse complemented?
        bool pandas.Series: would the extended seq be truncated by the ends of the subject sequence?
        bool pandas.Series: was the subject seq extended?
    """
    needs_revcomp = df.sstart > df.send
    add_to_end = df.qlen - df.qend
    add_to_start = df.qstart - 1
    ssum2 = (df.send + df.sstart) / 2.0
    sabs2 = np.abs(df.send - df.sstart) / 2.0
    end_idx = ssum2 + sabs2 - 1
    start_idx = ssum2 - sabs2 - 1
    start_idx[needs_revcomp] -= add_to_end
    start_idx[~needs_revcomp] -= add_to_start
    end_idx[needs_revcomp] += add_to_start
    end_idx[~needs_revcomp] += add_to_end
    clipped_start_idx = np.clip(start_idx, 0, (df.slen - 1))
    clipped_end_idx = np.clip(end_idx, 0, (df.slen - 1))
    trunc = (clipped_start_idx != start_idx) | (clipped_end_idx != end_idx)
    is_extended = (add_to_start > 0) | (add_to_end > 0)
    return clipped_start_idx, clipped_end_idx, needs_revcomp, trunc, is_extended


def retrieve_seq(seq, start, end, needs_revcomp):
    """Retrieve sub-sequence from nucleotide sequence

    Args:
        seq (str): nucleotide sequence
        start (int): Start index
        end (end): End index
        needs_revcomp (bool): reverse complement subseq?

    Returns:
         str: subseq of `seq`
    """
    start = int(start)
    end = int(end)
    out_seq = seq[start:(end + 1)]
    if needs_revcomp:
        out_seq = revcomp(out_seq)
    return out_seq


def parse_fasta(filepath):
    '''Parse a fasta file returning a generator yielding tuples of fasta headers to sequences.

    Note:
        This function should give equivalent results to SeqIO from BioPython

        .. code-block:: python

            from Bio import SeqIO
            # biopython to dict of header-seq
            hseqs_bio = {r.description:str(r.seq) for r in SeqIO.parse(fasta_path, 'fasta')}
            # this func to dict of header-seq
            hseqs = {header:seq for header, seq in parse_fasta(fasta_path)}
            # both methods should return the same dict
            assert hseqs == hseqs_bio

    Args:
        filepath (str): Fasta file path

    Returns:
        generator: yields tuples of (<fasta header>, <fasta sequence>)
    '''
    with open(filepath, 'r') as f:
        seqs = []
        header = ''
        line_count = 0
        for line in f:
            line = line.strip()
            if line == '':
                continue
            if line[0] == '>':
                if header == '':
                    header = line.replace('>', '')
                else:
                    yield header, ''.join(seqs)
                    seqs = []
                    header = line.replace('>', '')
            else:
                non_nucleotide_chars_in_line = set(line) - VALID_NUCLEOTIDES
                if len(non_nucleotide_chars_in_line) > 0:
                    msg = '{file}: Line {line} contains the following non-nucleotide characters: {chars}'.format(
                        file=filepath,
                        line=line_count,
                        chars=', '.join([x for x in non_nucleotide_chars_in_line]))
                    logging.warning(msg)
                seqs.append(line)
            line_count += 1
        yield header, ''.join(seqs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import argparse
import os
import sys
from typing import Optional

import attr
import logging
import re
import pandas as pd
from collections import defaultdict

from bio_hansel import program_name, program_desc, __version__
from bio_hansel.const import SUBTYPE_SUMMARY_COLS
from bio_hansel.subtyper import subtype_fasta, subtype_reads
from bio_hansel.subtype_stats import subtype_counts
from bio_hansel.subtyping_params import SubtypingParams
from bio_hansel.utils import genome_name_from_fasta_path, get_scheme_fasta, out_files_exists, get_scheme_params

SCRIPT_NAME = 'hansel'
LOG_FORMAT = '%(asctime)s %(levelname)s: %(message)s [in %(pathname)s:%(lineno)d]'


def init_console_logger(logging_verbosity=3):
    logging_levels = [logging.ERROR, logging.WARN, logging.INFO, logging.DEBUG]
    if logging_verbosity > (len(logging_levels) - 1):
        logging_verbosity = 3
    lvl = logging_levels[logging_verbosity]

    logging.basicConfig(format=LOG_FORMAT, level=lvl)


def init_parser():
    parser = argparse.ArgumentParser(prog=SCRIPT_NAME,
                                     formatter_class=argparse.RawDescriptionHelpFormatter,
                                     description=program_desc)
    parser.add_argument('files',
                        metavar='F',
                        nargs='*',
                        help='Input genome FASTA/FASTQ files')
    parser.add_argument('-s', '--scheme',
                        default='heidelberg',
                        help='Scheme to use for subtyping (built-in: "heidelberg", "enteritidis"; OR user-specified: /path/to/user/scheme)')
    parser.add_argument('--scheme-name',
                        help='Custom user-specified SNP substyping scheme name')
    parser.add_argument('-p', '--paired-reads',
                        nargs=2,
                        metavar=('forward_reads', 'reverse_reads'),
                        action='append',
                        help='FASTQ paired-end reads')
    parser.add_argument('-i', '--input-fasta-genome-name',
                        nargs=2,
                        metavar=('fasta_path', 'genome_name'),
                        action='append',
                        help='fasta file path to genome name pair')
    parser.add_argument('-D', '--input-directory',
                        help='directory of input fasta files (.fasta|.fa|.fna) or FASTQ files (paired FASTQ should have same basename with "_\d\.(fastq|fq)" postfix to be automatically paired)')
    parser.add_argument('-o', '--output-summary',
                        help='Subtyping summary output path (tab-delimited)')
    parser.add_argument('-O', '--output-tile-results',
                        help='Subtyping tile matching output path (tab-delimited)')
    parser.add_argument('-S', '--output-simple-summary',
                        help='Subtyping simple summary output path')
    parser.add_argument('--force',
                        action='store_true',
                        help='Force existing output files to be overwritten')
    parser.add_argument('--min-kmer-freq',
                        type=int,
                        help='Min k-mer freq/coverage')
    parser.add_argument('--max-kmer-freq',
                        type=int,
                        help='Max k-mer freq/coverage')
    # Changes
    parser.add_argument('--low-cov-depth-freq',
                        type=int,
                        help='Frequencies below this coverage are considered low coverage')
    parser.add_argument('--max-missing-tiles',
                        type=float,
                        help='Decimal proportion of maximum allowable missing tiles before being considered an error. (0.0 - 1.0)')
    parser.add_argument('--min-ambiguous-tiles',
                        type=int,
                        help='Minimum number of missing tiles to be considered an ambiguous result')
    parser.add_argument('--max-intermediate-tiles',
                        type=float,
                        help='Decimal proportion of maximum allowable missing tiles to be considered an intermediate subtype. (0.0 - 1.0)')
    # Changes
    parser.add_argument('-t', '--threads',
                        type=int,
                        default=1,
                        help='Number of parallel threads to run analysis (default=1)')
    parser.add_argument('-T', '--tmp-dir',
                        default='/tmp',
                        help='Base temporary working directory for intermediate analysis files')
    parser.add_argument('-K', '--keep-tmp',
                        default=False,
                        action='store_true',
                        help='Keep temporary analysis files')
    parser.add_argument('-v', '--verbose',
                        action='count',
                        default=0,
                        help='Logging verbosity level (-v == show warnings; -vvv == show debug info)')
    parser.add_argument('-V', '--version',
                        action='version',
                        version='%(prog)s {}'.format(__version__))
    return parser


def main():
    parser = init_parser()
    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()
    args = parser.parse_args()
    init_console_logger(args.verbose)
    output_summary_path = args.output_summary
    output_tile_results = args.output_tile_results
    output_simple_summary_path = args.output_simple_summary
    out_files_exists(output_simple_summary_path, args.force)
    out_files_exists(output_summary_path, args.force)
    out_files_exists(output_tile_results, args.force)
    scheme = args.scheme  # type: str
    scheme_name = args.scheme_name  # type: Optional[str]
    scheme_fasta = get_scheme_fasta(scheme)
    scheme_subtype_counts = subtype_counts(scheme_fasta)
    input_genomes = []
    reads = []
    logging.debug(args)

    subtyping_params = get_scheme_params(scheme)
    if not subtyping_params:
        subtyping_params = SubtypingParams()

    if args.low_cov_depth_freq:
        subtyping_params.low_coverage_depth_freq = args.low_cov_depth_freq
    if args.max_missing_tiles:
        subtyping_params.max_perc_missing_tiles = args.max_missing_tiles
    if args.min_ambiguous_tiles:
        subtyping_params.min_ambiguous_tiles = args.min_ambiguous_tiles
    if args.max_intermediate_tiles:
        subtyping_params.max_perc_intermediate_tiles = args.max_intermediate_tiles

    if args.files:
        fastas = [x for x in args.files if re.match(r'^.+\.(fasta|fa|fna)$', x)]
        fastqs = [x for x in args.files if re.match(r'^.+\.(fastq|fq)$', x)]
        if fastas:
            logging.info('# of input fastas %s', len(fastas))
            for fasta_path in fastas:
                fasta_path = os.path.abspath(fasta_path)
                if os.path.exists(fasta_path):
                    genome_name = genome_name_from_fasta_path(fasta_path)
                    input_genomes.append((fasta_path, genome_name))
                else:
                    logging.error('Input fasta "%s" does not exist!', fasta_path)
        if fastqs:
            logging.info('# of input fastqs %s', len(fastqs))
            grouped_fastqs = group_fastqs(fastqs)
            logging.info('Grouped %s fastqs into %s groups',
                         len(fastqs),
                         len(grouped_fastqs))
            reads += grouped_fastqs

    if args.input_fasta_genome_name:
        for fasta_path, genome_name in args.input_fasta_genome_name:
            input_genomes.append((os.path.abspath(fasta_path), genome_name))

    if args.input_directory:
        logging.info('Searching dir "%s" for FASTA files', args.input_directory)
        input_genomes += collect_fasta_from_dir(args.input_directory)
        logging.info('Searching dir "%s" for FASTQ files', args.input_directory)
        reads += collect_fastq_from_dir(args.input_directory)

    if args.paired_reads:
        for x in args.paired_reads:
            if not isinstance(x, (list, tuple)):
                logging.warning('Paired end reads not list or tuple %s', x)
                continue
            filenames = [os.path.basename(y) for y in x]
            common_prefix = os.path.commonprefix(filenames)
            genome_name = re.sub(r'[\W\_]+$', r'', common_prefix)
            if genome_name == '':
                genome_name = filenames[0]
            reads.append((x, genome_name))

    if len(input_genomes) == 0 and len(reads) == 0:
        raise Exception('No input files specified!')

    n_threads = args.threads
    tmp_dir = args.tmp_dir

    subtype_results = []
    dfs = []
    if input_genomes:
        if n_threads == 1:
            logging.info('Serial single threaded run mode on %s input genomes', len(input_genomes))
            outputs = [subtype_fasta(subtyping_params,
                                     scheme,
                                     input_fasta,
                                     genome_name,
                                     tmp_dir=tmp_dir,
                                     scheme_name=scheme_name,
                                     scheme_subtype_counts=scheme_subtype_counts)
                       for input_fasta, genome_name in input_genomes]
        else:
            from multiprocessing import Pool
            logging.info('Initializing thread pool with %s threads', n_threads)
            pool = Pool(processes=n_threads)
            logging.info('Running analysis asynchronously on %s input genomes', len(input_genomes))
            res = [pool.apply_async(subtype_fasta, (subtyping_params,
                                                    scheme,
                                                    input_fasta,
                                                    genome_name,
                                                    tmp_dir,
                                                    scheme_name,
                                                    scheme_subtype_counts))
                   for input_fasta, genome_name in input_genomes]

            logging.info('Parallel analysis complete! Retrieving analysis results')
            outputs = [x.get() for x in res]

        for subtype, df in outputs:
            if df is not None:
                dfs.append(df)
            subtype_results.append(attr.asdict(subtype))

    if reads:
        outputs = [subtype_reads(subtyping_params,
                                 scheme=scheme,
                                 reads=r,
                                 genome_name=genome_name,
                                 tmp_dir=tmp_dir,
                                 threads=n_threads,
                                 scheme_name=scheme_name,
                                 scheme_subtype_counts=scheme_subtype_counts)
                   for r, genome_name in reads]

        for subtype, df in outputs:
            if df is not None:
                dfs.append(df)
            subtype_results.append(attr.asdict(subtype))

    dfall = pd.concat(dfs)  # type: pd.DataFrame
    dfsummary = pd.DataFrame(subtype_results)
    dfsummary = dfsummary[SUBTYPE_SUMMARY_COLS]

    df_simple_summary = dfsummary[['sample', 'subtype', 'qc_status', 'qc_message']]

    if output_summary_path:
        dfsummary.to_csv(output_summary_path, sep='\t', index=None)
        logging.info('Wrote subtyping output summary to %s', output_summary_path)
    else:
        print(dfsummary.to_csv(sep='\t', index=None))

    if output_tile_results:
        dfall.to_csv(output_tile_results, sep='\t', index=None)

    if output_simple_summary_path:
        df_simple_summary.to_csv(output_simple_summary_path, sep='\t', index=None)


def collect_fasta_from_dir(input_directory):
    input_genomes = []
    for x in os.listdir(input_directory):
        full_file_path = os.path.abspath(os.path.join(input_directory, x))
        if os.path.isfile(full_file_path) and re.match(r'^.+\.(fasta|fa|fna)$', x):
            genome_name = genome_name_from_fasta_path(full_file_path)
            input_genomes.append((full_file_path, genome_name))
    return input_genomes


def group_fastqs(fastqs):
    reads = []
    genome_fastqs = defaultdict(list)
    for fastq in fastqs:
        filename = os.path.basename(fastq)
        basefilename = re.sub(r'_\d\.(fastq|fq)$', '', filename)
        genome_fastqs[basefilename].append(fastq)
    for genome_name, fastq_paths in genome_fastqs.items():
        reads.append((fastq_paths, genome_name))
    return reads


def collect_fastq_from_dir(input_directory):
    fastqs = []
    for x in os.listdir(input_directory):
        full_file_path = os.path.abspath(os.path.join(input_directory, x))
        if os.path.isfile(full_file_path) and re.match(r'^.+\.(fastq|fq)$', x):
            fastqs.append(full_file_path)
    if len(fastqs) > 0:
        logging.info('Found %s FASTQ files in %s',
                     len(fastqs),
                     input_directory)
        reads_from_dir = group_fastqs(fastqs)
        logging.info('Collected %s read sets from %s FASTQ files in %s',
                     len(reads_from_dir),
                     len(fastqs),
                     input_directory)
        return reads_from_dir
    return []


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

import attr
from datetime import datetime
import logging
import shutil
import os
import pandas as pd
import numpy as np
from pandas.io.common import EmptyDataError
import re
from .const import BLAST_TABLE_COLS
from ..utils import exc_exists, run_command


@attr.s
class BlastRunner:
    fasta_path = attr.ib()
    tmp_work_dir = attr.ib(default='/tmp', validator=attr.validators.instance_of(str))
    blast_db_created = attr.ib(default=False, validator=attr.validators.instance_of(bool))
    makeblastdb_exc = attr.ib(default='makeblastdb', validator=attr.validators.instance_of(str))
    blastn_exc = attr.ib(default='blastn', validator=attr.validators.instance_of(str))

    @makeblastdb_exc.validator
    def _check_makeblastdb_exists(self, attribute, value):
        if not exc_exists(value):
            raise OSError('makeblastdb executable "{}" does not exist in the user $PATH'.format(value))

    @blastn_exc.validator
    def _check_blastn_exists(self, attribute, value):
        if not exc_exists(value):
            raise OSError('blast executable "{}" does not exist in the user $PATH'.format(value))

    @fasta_path.validator
    def _fasta_path_exists(self, attribute, value):
        if not os.path.exists(value):
            raise OSError('FASTA file does not exist at {}'.format(value))

    def _create_tmp_folder(self):
        count = 1
        tmp_dir = self.tmp_work_dir
        while True:
            try:
                logging.info('Trying to create analysis directory at: %s', tmp_dir)
                os.makedirs(tmp_dir)
                break
            except OSError as e:
                logging.warning('Error on creation of tmp analysis directory "{}"! {}'.format(
                    tmp_dir,
                    e
                ))
                tmp_dir = '{}_{}'.format(self.tmp_work_dir, count)
                count += 1
        self.tmp_work_dir = tmp_dir
        return self.tmp_work_dir

    def _copy_fasta_to_work_dir(self):
        filename = os.path.basename(self.fasta_path)
        filename, ext = os.path.splitext(filename)
        filename_no_spaces = re.sub(r'\W', '_', filename)
        dest_path = os.path.join(self.tmp_work_dir, filename_no_spaces + ext)
        if self.fasta_path == dest_path:
            self.tmp_fasta_path = dest_path
            return dest_path
        shutil.copyfile(self.fasta_path, dest_path)
        self.tmp_fasta_path = dest_path
        return dest_path

    def _run_makeblastdb(self):
        work_dir = os.path.dirname(self.tmp_fasta_path)
        filename = os.path.basename(self.tmp_fasta_path)
        nin_filepath = os.path.join(work_dir, filename + '.nin')
        if os.path.exists(nin_filepath):
            self.blast_db_created = True
            return self.tmp_fasta_path
        cmdlist = [self.makeblastdb_exc,
                   '-in', '{}'.format(self.tmp_fasta_path),
                   '-dbtype', 'nucl']
        exit_code, stdout, stderr = run_command(cmdlist)
        if exit_code != 0:
            raise Exception('Error {}: makeblastdb could not create a BLAST DB for {}. stderr: {}'.format(exit_code,
                                                                                                          self.tmp_fasta_path,
                                                                                                          stderr))
        if stdout is not None and stdout != '':
            logging.debug('makeblastdb on {0} STDOUT: {1}'.format(self.tmp_fasta_path, stdout))
        if stderr is not None and stderr != '':
            logging.debug('makeblastdb on {0} STDERR: {1}'.format(self.tmp_fasta_path, stderr))
        if os.path.exists(nin_filepath):
            self.blast_db_created = True
            return self.tmp_fasta_path
        else:
            ex_msg = 'makeblastdb was not able to create a BLAST DB for {0}. STDERR: {1}'.format(filename, stderr)
            logging.error(ex_msg)
            raise Exception(ex_msg)

    def blast_against_query(self, query_fasta_path, blast_task='megablast', evalue=1e-4, min_pid=85, word_size=22):
        if not self.blast_db_created:
            self.prep_blast()

        query_filename = os.path.basename(query_fasta_path)
        db_filename = os.path.basename(self.tmp_fasta_path)
        timestamp = '{:%Y%b%d_%H_%M_%S}'.format(datetime.now())
        outfile = os.path.join(self.tmp_work_dir, '{}-{}-{}.blast'.format(query_filename,
                                                                          db_filename,
                                                                          timestamp))
        cmd_list = [self.blastn_exc,
                    '-task', blast_task,
                    '-query', query_fasta_path,
                    '-db', '{}'.format(self.tmp_fasta_path),
                    '-word_size', '{}'.format(word_size),
                    '-evalue', '{}'.format(evalue),
                    '-dust', 'no',
                    '-perc_identity', '{}'.format(min_pid),
                    '-out', outfile,
                    '-outfmt', '6 {}'.format(' '.join(BLAST_TABLE_COLS))]
        logging.info('Running commandline "{}"'.format(' '.join(cmd_list)))
        exit_code, stdout, stderr = run_command(cmd_list)
        if os.path.exists(outfile):
            return outfile
        else:
            err_msg_fmt = 'error code {}: blastn on db {} and query {} did not produce expected output file at {}. stderr: {}'
            ex_msg = err_msg_fmt.format(exit_code, db_filename, query_filename, outfile, stderr)
            raise Exception(ex_msg)

    def cleanup(self):
        self.blast_db_created = False
        shutil.rmtree(self.tmp_work_dir)

    def prep_blast(self):
        self._create_tmp_folder()
        self._copy_fasta_to_work_dir()
        self._run_makeblastdb()

    def __enter__(self):
        if self.blast_db_created:
            return self
        self.prep_blast()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.cleanup()


@attr.s
class BlastReader:
    blast_outfile = attr.ib(validator=attr.validators.instance_of(str))
    df = attr.ib(default=None, validator=attr.validators.optional(attr.validators.instance_of(pd.DataFrame)))

    def parse(self):
        """Parse tabular blastn output file into a pandas DataFrame

        Sort the DataFrame by BLAST bitscore, compute query coverage and if result is truncated by end of subject
        sequence/contig.

        Returns:
            pandas.DataFrame: dataframe of tabular BLAST results
            None: if no results could be parsed from BLAST output file

        Exceptions:
            EmptyDataError: No data could be parsed from the `blastn` output file
        """
        try:
            self.df = pd.read_table(self.blast_outfile, header=None)
            self.df.columns = BLAST_TABLE_COLS
            # calculate the coverage for when results need to be validated
            self.df.loc[:, 'coverage'] = self.df.length / self.df.qlen
            self.df.sort_values(by='bitscore', ascending=False, inplace=True)
            self.df.loc[:, 'is_trunc'] = BlastReader.trunc(qstart=self.df.qstart,
                                                           qend=self.df.qend,
                                                           qlen=self.df.qlen,
                                                           sstart=self.df.sstart,
                                                           send=self.df.send,
                                                           slen=self.df.slen)

            return self.df

        except EmptyDataError as exc:
            logging.warning('No BLASTN results to parse from file %s', self.blast_outfile)
            return None

    def to_dict(self):
        if self.df is not None:
            return self.df.to_dict()
        else:
            return None

    @staticmethod
    def trunc(qstart, qend, sstart, send, qlen, slen):
        """Check if a query sequence is truncated by the end of a subject sequence

        Args:
            qstart (int pandas.Series): Query sequence start index
            qend (int pandas.Series): Query sequence end index
            sstart (int pandas.Series): Subject sequence start index
            send (int pandas.Series): Subject sequence end index
            qlen (int pandas.Series): Query sequence length
            slen (int pandas.Series): Subject sequence length

        Returns:
            Boolean pandas.Series: Result truncated by subject sequence end?
        """
        ssum2 = (send + sstart) / 2.0
        sabs2 = np.abs(send - sstart) / 2.0
        smax = ssum2 + sabs2
        smin = ssum2 - sabs2
        q_match_len = np.abs(qstart - qend) + 1
        return (q_match_len < qlen) & ((smax >= slen) | (smin <= 1))

    def perfect_matches(self):
        """
        Return pandas DataFrame with perfect BLAST matches (100% identity and coverage)

        Returns:
            pandas.DataFrame or None: DataFrame of perfect BLAST matches or None if no perfect matches exist
        """
        if self.df is None:
            return None

        df_perfect_matches = self.df[(self.df['coverage'] == 1.0) & (self.df['pident'] == 100.0)]
        if df_perfect_matches.shape[0] == 0:
            return None
        return df_perfect_matches

    def __enter__(self):
        if self.blast_outfile is None:
            return self
        self.parse()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.df = None
        self.blast_outfile = None



# -*- coding: utf-8 -*-

import logging
import os
import re
from datetime import datetime
from typing import Optional, List, Dict, Union, TYPE_CHECKING, Tuple

from pandas import DataFrame

if TYPE_CHECKING:
    from .subtype_stats import SubtypeCounts

from . import program_name
from .quality_check import perform_quality_check
from .blast_wrapper import BlastRunner, BlastReader
from .kmer_count import Jellyfisher
from .subtype import Subtype
from .subtyping_params import SubtypingParams
from .utils import find_inconsistent_subtypes, get_scheme_fasta, get_scheme_version
from .subtype_stats import subtype_counts
from .const import FASTA_COLUMNS_TO_REMOVE


def subtype_fasta(subtyping_params: SubtypingParams,
                  scheme: str,
                  fasta_path: str,
                  genome_name: str,
                  tmp_dir: str = '/tmp',
                  scheme_name: Optional[str] = None,
                  scheme_subtype_counts: Optional[Dict[str, 'SubtypeCounts']] = None) -> (Subtype, DataFrame):
    dtnow = datetime.now()
    genome_name_no_spaces = re.sub(r'\W', '_', genome_name)
    genome_tmp_dir = os.path.join(tmp_dir,
                                  dtnow.strftime("%Y%m%d%H%M%S") + '-' + program_name + '-' + genome_name_no_spaces)
    scheme_fasta = get_scheme_fasta(scheme)
    if scheme_subtype_counts is None:
        scheme_subtype_counts = subtype_counts(scheme_fasta)
    scheme_version = get_scheme_version(scheme)
    with BlastRunner(fasta_path=fasta_path, tmp_work_dir=genome_tmp_dir) as brunner:
        blast_outfile = brunner.blast_against_query(scheme_fasta, word_size=33)
        with BlastReader(blast_outfile=blast_outfile) as breader:
            df = breader.parse()

    st = Subtype(sample=genome_name, file_path=fasta_path, scheme=scheme_name or scheme, scheme_version=scheme_version, scheme_subtype_counts=scheme_subtype_counts)
    if df is None or df.shape[0] == 0:
        logging.warning('No Heidelberg subtyping tile matches for "%s"', fasta_path)
        st.are_subtypes_consistent = False
        return st, None

    df.rename(columns={'qseqid': 'tilename',
                       'sseq': 'seq'},
              inplace=True)

    refpositions = [x for x, y in df.tilename.str.split('-')]
    subtypes = [y for x, y in df.tilename.str.split('-')]
    df['refposition'] = [int(x.replace('negative', '')) for x in refpositions]
    df['subtype'] = subtypes
    df['is_pos_tile'] = ~df.tilename.str.contains('negative')
    logging.debug('df: %s', df)
    dfpos = df[df.is_pos_tile]
    logging.debug('dfpos: %s', dfpos)
    subtype_lens = dfpos.subtype.apply(len)
    max_subtype_strlen = subtype_lens.max()
    logging.debug('max substype str len: %s', max_subtype_strlen)
    dfpos_highest_res = dfpos[subtype_lens == max_subtype_strlen]
    logging.debug('dfpos_highest_res: %s', dfpos_highest_res)
    pos_subtypes = [[int(y) for y in x.split('.')] for x in dfpos.subtype.unique()]
    pos_subtypes.sort(key=lambda a: len(a))
    logging.debug('pos_subtypes: %s', pos_subtypes)
    inconsistent_subtypes = find_inconsistent_subtypes(pos_subtypes)
    logging.debug('inconsistent_subtypes: %s', inconsistent_subtypes)
    st.n_tiles_matching_all = df.tilename.unique().size
    st.n_tiles_matching_positive = dfpos.tilename.unique().size
    st.n_tiles_matching_subtype = dfpos_highest_res.tilename.unique().size
    pos_subtypes_str = [x for x in dfpos.subtype.unique()]
    pos_subtypes_str.sort(key=lambda x: len(x))
    st.all_subtypes = '; '.join(pos_subtypes_str)
    subtype_list = [x for x in dfpos_highest_res.subtype.unique()]
    st.subtype = '; '.join(subtype_list)
    st.n_tiles_matching_all_expected = ';'.join([str(scheme_subtype_counts[x].all_tile_count) for x in subtype_list])
    st.n_tiles_matching_positive_expected = ';'.join(
        [str(scheme_subtype_counts[x].positive_tile_count) for x in subtype_list])
    st.n_tiles_matching_subtype_expected = ';'.join(
        [str(scheme_subtype_counts[x].subtype_tile_count) for x in subtype_list])
    st.tiles_matching_subtype = '; '.join([x for x in dfpos_highest_res.tilename.unique()])

    if len(inconsistent_subtypes) > 0:
        st.are_subtypes_consistent = False
        st.inconsistent_subtypes = inconsistent_subtypes

    possible_downstream_subtypes = [s for s in scheme_subtype_counts
                                       if re.search("^({})(\.)(\d)$".format(re.escape(st.subtype)), s)]
    non_present_subtypes = []
    if possible_downstream_subtypes:
        for subtype in possible_downstream_subtypes:
            blah = df['subtype']
            # wtf is going on
            if not any(df.subtype == subtype):
                non_present_subtypes.append(subtype)

    st.non_present_subtypes = non_present_subtypes

    perform_quality_check(st, df, subtyping_params)

    logging.info(st)

    df['sample'] = genome_name
    df['file_path'] = fasta_path
    df['scheme'] = scheme_name or scheme
    df['scheme_version'] = scheme_version
    df['qc_status'] = st.qc_status
    df['qc_message'] = st.qc_message

    df = df[df.columns[~df.columns.isin(FASTA_COLUMNS_TO_REMOVE)]]
    return st, df


def subtype_reads(subtyping_params: SubtypingParams,
                  scheme: str,
                  reads: Union[str, List[str]],
                  genome_name: str,
                  tmp_dir: str = '/tmp',
                  threads: int = 1,
                  scheme_name: Optional[str] = None,
                  scheme_subtype_counts: Optional[Dict[str, 'SubtypeCounts']] = None) -> Tuple[Subtype, DataFrame]:
    dtnow = datetime.now()
    genome_name_no_spaces = re.sub(r'\W', '_', genome_name)
    genome_tmp_dir = os.path.join(tmp_dir,
                                  dtnow.strftime("%Y%m%d%H%M%S") + '-' + program_name + '-' + genome_name_no_spaces)
    scheme_fasta = get_scheme_fasta(scheme)
    if scheme_subtype_counts is None:
        scheme_subtype_counts = subtype_counts(scheme_fasta)
    scheme_version = get_scheme_version(scheme)
    with Jellyfisher(scheme=scheme_name or scheme,
                     scheme_version=scheme_version,
                     scheme_subtype_counts=scheme_subtype_counts,
                     scheme_fasta=scheme_fasta,
                     genome_name=genome_name,
                     reads=reads,
                     min_kmer_freq=subtyping_params.min_kmer_freq,
                     max_kmer_freq=subtyping_params.max_kmer_freq,
                     tmp_dir=genome_tmp_dir,
                     threads=threads) as jfer:
        st, df = jfer.summary()

        perform_quality_check(st, df, subtyping_params)

        df['scheme'] = scheme_name or scheme
        df['scheme_version'] = scheme_version
        df['qc_status'] = st.qc_status
        df['qc_message'] = st.qc_message

        df = df[df.columns[~df.columns.isin(FASTA_COLUMNS_TO_REMOVE)]]

        return st, df



from pandas import DataFrame

from ..subtyping_params import SubtypingParams
from ..quality_check.const import FAIL_MESSAGE, WARNING_MESSAGE, MISSING_TILES_ERROR_1, \
    MIXED_SAMPLE_ERROR_2, AMBIGUOUS_RESULTS_ERROR_3, INTERMEDIATE_SUBTYPE_WARNING, \
    NON_CONFIDENT_RESULTS_ERROR_4
from ..quality_check.qc_utils import get_conflicting_tiles, get_num_pos_neg_tiles
from ..subtype import Subtype
from typing import Tuple, Optional
import logging


def does_subtype_result_exist(st: Subtype) -> bool:
    """ Check if the subtype result exists.
    Note:
            This method verifies that there is an expected matching subtype so we can proceed
            with quality checking.

    Input:
            :param st: Subtyping results.

    Returns:
            Bool: True if the subtype exists
            Bool: False if the subtype does not exist
    """
    logging.debug("QC: Checking if subtype result exists.")
    return st.subtype is not None and 0 < len(st.subtype)


def check_missing_tiles(st: Subtype, df: DataFrame, p: SubtypingParams) -> Tuple[Optional[str], Optional[str]]:
    """ Check if there's more than 5% missing tiles.
    Note:
            check_missing_tiles will check if more than 5% of the scheme's tiles are missing.
            If they are, we calculate the average frequency coverage depth for the tiles that are present, and
            provide an adequate error message based on the coverage.

    Args:
            :param st: Subtyping results.
            :param df: DataFrame containing subtyping results.
            :param p: Subtyping parameters for quality check thresholds.

    Returns:
            Tuple[Optional[str], Optional[str]]
            error_status: Contains any error statuses { Warning, Fail }
            error_messages: Contains the reasoning for any error status.
    """
    logging.debug("QC: Checking for missing tiles.")
    error_status = None
    error_messages = None

    if not st.are_subtypes_consistent:
        logging.debug("QC: Checking for missing tiles not run, inconsistent subtype detected.")
        error_messages = "Subtype is inconsistent, quality checking for missing tiles not run."
        error_status = FAIL_MESSAGE
        return error_status, error_messages

    exp = int(st.n_tiles_matching_all_expected)
    obs = int(st.n_tiles_matching_all)

    if (exp - obs) / exp > p.max_perc_missing_tiles:
        tiles_with_hits = df[df['is_kmer_freq_okay']] # type: DataFrame
        avg_depth = tiles_with_hits['freq'].mean()

        error_messages = "{}: More than {:.2%} missing tiles were detected. {} Avg calculated tile coverage = {}".format(
            MISSING_TILES_ERROR_1, p.max_perc_missing_tiles,
            "Low coverage detected, possibly need more whole genome sequencing data."
            if avg_depth < p.low_coverage_depth_freq
            else "Adequate coverage detected, this may be the wrong serovar/species for scheme: {}".format(st.scheme),
            avg_depth
        )

        error_status = FAIL_MESSAGE

    return error_status, error_messages


def check_mixed_subtype(st: Subtype, df: DataFrame, p: SubtypingParams) -> Tuple[Optional[str], Optional[str]]:
    """ Check to see if the subtype is mixed.
    Note:
            check_mixed_subtype will check the results of the typing for:
            1) + and - tiles that are present for the target site above minimum coverage threshold.
            2) If the subtyping result came back with an inconsistent call.
            Then the method will provide an adequate error message to the user.

    Args:
            :param st: Subtyping results.
            :param df: DataFrame containing subtyping results.
            :param p: Subtyping parameters for quality check thresholds.

    Returns:
            Tuple[Optional[str], Optional[str]]
            error_status: Contains any error statuses { Warning, Fail }
            error_messages: Contains the reasoning for any error status.
    """
    logging.debug("QC: Checking for mixed subtypes.")
    error_status = None
    error_messages = None

    # First check the obvious, if the st is consistent, of course it's a mixed subtype.
    if not st.are_subtypes_consistent or st.inconsistent_subtypes:
        error_status = FAIL_MESSAGE
        error_messages = "{}: Mixed subtypes detected. "\
                         "Mixed subtypes found: {}.".format(MIXED_SAMPLE_ERROR_2, st.inconsistent_subtypes)
    else:
        # Find the conflicting tiles
        conflicting_tiles = get_conflicting_tiles(st, df)
        if 0 < conflicting_tiles.shape[0]:
            error_status = FAIL_MESSAGE
            error_messages = "{}: Mixed subtype detected." \
                             " Positive and negative tiles detected for the same target site" \
                             " {} for subtype {}.".format(MIXED_SAMPLE_ERROR_2,
                                                          conflicting_tiles['refposition'].tolist(), st.subtype)

    return error_status, error_messages


def is_missing_target_sites(st: Subtype, df: DataFrame, p: SubtypingParams) -> Tuple[Optional[str], Optional[str]]:
    """Are there X or more missing target sites for an expected subtype? (ERROR 3)
    Note:
            This method will check if there are any refpositions missing from the subtyping scheme in the result.
            If there are missing refpositions, this is a missing target. If there are too many missing targets,
            this result is an Ambiguous Result.
    Args:
            :param st: Subtyping results.
            :param df: DataFrame containing subtyping results.
            :param p: Subtyping parameters for quality check thresholds.

    Returns:
            Tuple[Optional[str], Optional[str]]
            error_status: Contains any error statuses { Warning, Fail }
            error_messages: Contains the reasoning for any error status.
    """
    logging.debug("QC: Checking for ambiguous results.")
    error_status = None
    error_messages = None

    # We do this check as inconsistent subtypes fall within mixed subtypes. This checks for inconsistent results.
    if not st.are_subtypes_consistent:
        logging.debug("QC: Checking for ambiguous results not run, inconsistent subtype detected.")
        error_messages = "Subtype is inconsistent, quality checking for ambiguous results not run."
        error_status = FAIL_MESSAGE
        return error_status, error_messages

    potential_subtypes = st.all_subtypes.split('; ')
    uniq_positions = {y for x in potential_subtypes for y in st.scheme_subtype_counts[x].refpositions}
    missing_targets = [x for x in uniq_positions if (df.refposition == x).sum() == 0]

    exp = int(st.n_tiles_matching_all_expected)
    obs = int(st.n_tiles_matching_all)
    if (exp - obs) / exp <= p.max_perc_missing_tiles and len(missing_targets) >= p.min_ambiguous_tiles:
            error_status = FAIL_MESSAGE
            error_messages = ("{}: {} missing positions detected for subtype: {}."
                              .format(
                                        AMBIGUOUS_RESULTS_ERROR_3,
                                        len(missing_targets),
                                        st.subtype
                                     )
                              )

    return error_status, error_messages


def is_missing_downstream_targets(st: Subtype, df: DataFrame, p: SubtypingParams) -> Tuple[Optional[str], Optional[str]]:
    """ Check for missing downstream targets (ERROR 4)
    Note:
            This method will check if there's any non_present_subtypes in the result, which would indicate a non confident
            result. This is due to the fact if you have a subtyping result of `2.1.1.2` and you're missing `2.1.1.2.X`
            You can't be sure that the subtype's final call is 2.1.1.2 if you're missing information.

    Args:
            :param st: Subtyping results.

    Output:
            Tuple[Optional[str], Optional[str]]
            error_status: Contains any error statuses { FAIL_MESSAGE }
            error_messages: Contains the reasoning for any fail status.
    """
    error_status = None
    error_messages = None

    non_present_subtypes = st.non_present_subtypes
    if non_present_subtypes:
        error_status = FAIL_MESSAGE
        error_messages = ("{}: Subtype {} was found, but downstream subtype(s) {} tiles were missing."
                          " Therefore due to the missing downstream tiles, there is a lack of confidence in the"
                          "final subtype call."
                          .format(
                            NON_CONFIDENT_RESULTS_ERROR_4,
                            st.subtype,
                            non_present_subtypes
                           )
                          )

    return error_status, error_messages


def check_intermediate_subtype(st: Subtype, df: DataFrame, p: SubtypingParams) -> Tuple[Optional[str], Optional[str]]:
    """ Check for intermediate subtypes within the result.
    Note:
            Check_intermediate_subtype will return a WARNING message if 95% of the scheme tiles are found, there are
            0 contradicting tiles (tiles with same position but there exists + and -), total subtype tiles < expected,
            and +/- tiles exist in the result.

    Args:
            :param st: Subtyping results.
            :param df: DataFrame containing subtyping results.
            :param p: Subtyping parameters for quality check thresholds.

    Output:
            Tuple[Optional[str], Optional[str]]
            error_status: Contains any error statuses { Warning }
            error_messages: Contains the reasoning for any warning status.
    """
    logging.debug("QC: Checking for intermediate subtypes.")
    error_status = None
    error_messages = None

    if not st.are_subtypes_consistent:
        logging.debug("QC: Checking for intermediate subtypes not run, inconsistent subtype detected.")
        error_messages = "Subtype is inconsistent, quality checking for intermediate subtype not run."
        error_status = FAIL_MESSAGE
        return error_status, error_messages

    total_subtype_tiles = int(st.n_tiles_matching_subtype_expected)
    total_subtype_tiles_hits = int(st.n_tiles_matching_subtype)
    conflicting_tiles = get_conflicting_tiles(st, df)
    num_pos_tiles, num_neg_tiles = get_num_pos_neg_tiles(st, df)

    obs = int(st.n_tiles_matching_all)
    exp = int(st.n_tiles_matching_all_expected)

    if (exp - obs) / exp <= p.max_perc_intermediate_tiles and conflicting_tiles.shape[0] == 0 and \
            total_subtype_tiles_hits < total_subtype_tiles and num_pos_tiles and num_neg_tiles:

        error_status = WARNING_MESSAGE
        error_messages = "{}: Possible intermediate subtype. All scheme tiles were found, " \
                         "but a fraction were positive for the final subtype. " \
                         "Total subtype hits: {} | Total subtype expected: {}." \
                         .format(INTERMEDIATE_SUBTYPE_WARNING, total_subtype_tiles_hits, total_subtype_tiles)

    return error_status, error_messages



from typing import Tuple

from pandas import DataFrame

from ..subtype import Subtype


def get_conflicting_tiles(st: Subtype, df: DataFrame) -> list:
    """ This method gets positive and negative tiles that both are present for a subtype.
    Note:
            The purpose of this method is to find positive and negative tiles for the same refposition in the DataFrame.
            The method will return a list with the conflicting tiles.

    Args:
            :param st: Subtyping results.
            :param df: DataFrame containing subtyping results.

    Returns:
            DataFrame containing the conflicting positive and negative tiles.
    """
    if st.subtype:
        if 'is_kmer_freq_okay' in df:
            dfst = df[(df['subtype'] == str(st.subtype)) & (df['is_kmer_freq_okay'])]
        else:  # fasta files
            dfst = df[(df['subtype'] == str(st.subtype))]

    pos_tile_positions = dfst[dfst['is_pos_tile']]['refposition'].tolist()
    neg_tiles = dfst[~dfst['is_pos_tile']]
    conflicting_tiles = neg_tiles[neg_tiles['refposition'].isin(pos_tile_positions)]

    return conflicting_tiles


def get_num_pos_neg_tiles(st: Subtype, df: DataFrame) -> Tuple[int, int]:
    """ This method gets the number of positive and negative tiles.
    Note:
            The purpose of this method is to find the count of positive and negative tiles, and return them to the
            caller.

    Args:
            :param st: Subtyping results.
            :param df: DataFrame containing subtyping results.

    Returns:
            Tuple[int,int] containing the count of positive and negative tiles.
    """
    num_pos_tiles = 0
    num_neg_tiles = 0

    if st.subtype:
        dfst = df[(df['subtype'] == str(st.subtype))]
        num_pos_tiles = dfst[dfst['is_pos_tile']].shape[0]
        num_neg_tiles = dfst[~dfst['is_pos_tile']].shape[0]

    return num_pos_tiles, num_neg_tiles





FAIL_MESSAGE = "FAIL"
WARNING_MESSAGE = "WARNING"
# Errors for Hansel
MISSING_TILES_ERROR_1 = "Missing Tiles Error 1"
MIXED_SAMPLE_ERROR_2 = "Mixed Sample Error 2"
AMBIGUOUS_RESULTS_ERROR_3 = "Ambiguous Results Error 3"
NON_CONFIDENT_RESULTS_ERROR_4 = "Non Confident Results Error 4"
INTERMEDIATE_SUBTYPE_WARNING = "Intermediate Subtype Warning"



import logging
import os
import re
from subprocess import Popen, PIPE
from typing import List, Any, Optional

from .const import SCHEME_FASTAS
from .subtyping_params import SubtypingParams


def run_command(cmdlist: List[str]) -> (int, str, str):
    p = Popen(cmdlist,
              stdout=PIPE,
              stderr=PIPE)
    exit_code = p.wait()
    stdout, stderr = p.communicate()
    if isinstance(stdout, bytes):
        stdout = stdout.decode()
    if isinstance(stderr, bytes):
        stderr = stderr.decode()
    return exit_code, stdout, stderr


def exc_exists(exc_name: str) -> bool:
    """Check if an executable exists

    Args:
        exc_name (str): Executable name or path (e.g. "blastn")

    Returns:
        bool: Does the executable exists in the user's $PATH?
    """
    cmd = ['which', exc_name]
    exit_code, stdout, stderr = run_command(cmd)
    if exit_code == 0:
        return True
    else:
        logging.warning('which exited with non-zero code {} with command "{}"'.format(exit_code, ' '.join(cmd)))
        logging.warning(stderr)
        return False


def out_files_exists(filepath: str, force: bool):
    if filepath:
        file_exists_err_fmt = 'File "{}" already exists! If you want to overwrite this output file run with opt "--force"'
        if os.path.exists(filepath):
            if not force:
                raise OSError(file_exists_err_fmt.format(filepath))
            else:
                logging.warning('File "{}" already exists, overwriting with "--force" - uh oh :S')


def genome_name_from_fasta_path(fasta_path: str) -> str:
    """Extract genome name from fasta filename

    Get the filename without directory and remove the file extension.

    Example:
        With fasta file path ``/path/to/genome_1.fasta``::

            fasta_path = '/path/to/genome_1.fasta'
            genome_name = genome_name_from_fasta_path(fasta_path)
            print(genome_name)
            # => "genome_1"

    Args:
        fasta_path (str): fasta file path

    Returns:
        str: genome name
    """
    filename = os.path.basename(fasta_path)
    return re.sub(r'(\.fa$)|(\.fas$)|(\.fasta$)|(\.fna$)|(\.\w{1,}$)', '', filename)


def compare_subtypes(a: List[Any], b: List[Any]) -> bool:
    for x, y in zip(a, b):
        if x != y:
            return False
    return True


def find_inconsistent_subtypes(subtypes: List[Any]) -> List[str]:
    from collections import Counter
    incon = []
    for i in range(len(subtypes) - 1):
        a = subtypes[i]
        for j in range(i + 1, len(subtypes)):
            b = subtypes[j]
            is_consistent = compare_subtypes(a, b)
            if not is_consistent:
                incon.append((a, b))
    l = []
    for a, b in incon:
        astr = '.'.join([str(x) for x in a])
        bstr = '.'.join([str(x) for x in b])
        l += [astr, bstr]
    c = Counter(l)
    incon_subtypes = []
    for subtype, freq in c.most_common():
        if freq >= 1:
            incon_subtypes.append(subtype)
        else:
            break
    return incon_subtypes


def get_scheme_fasta(scheme: str) -> str:
    if scheme in SCHEME_FASTAS:
        scheme_fasta = SCHEME_FASTAS[scheme]['file']
    elif os.path.exists(scheme) and os.path.isfile(scheme):
        scheme_fasta = scheme
    else:
        raise FileNotFoundError('Could not find user-specified subtyping scheme fasta "%s"', scheme)
    return scheme_fasta


def get_scheme_params(scheme: str) -> Optional[SubtypingParams]:
    if scheme in SCHEME_FASTAS:
        return SCHEME_FASTAS[scheme]['subtyping_params']


def get_scheme_version(scheme: str) -> Optional[str]:
    if scheme in SCHEME_FASTAS:
        version = SCHEME_FASTAS[scheme]['version']  # type: str
        return version
    return None
